%
%     Implementation by Piotr Hoownia <holownia@student.agh.edu.pl>
%
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     PlNXT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of PlNXT.
%
%     PlNXT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     PlNXT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with PlNXT.  If not, see <http://www.gnu.org/licenses/>.
%

:- module(nxt_actions_simulator,[
				nxt_actions_connection_open/0,
				nxt_actions_connection_close/0,
				nxt_actions_motor/3,
				nxt_actions_motor_is_moving/1,
				nxt_actions_motor_stop/1,
				nxt_actions_motor_stop/2,
				nxt_actions_motor_get_speed/2,
				nxt_actions_motor_get_rotations/2,
				nxt_actions_motor_reset_position/2,
				nxt_actions_touch_sensor/2,
				nxt_actions_sound_sensor/2,
				nxt_actions_light_sensor/2,
				nxt_actions_light_sensor_LED/2,
				nxt_actions_ultrasonic_sensor/2,
				nxt_actions_voltage/1,
				nxt_actions_voltage_millivolt/1,
				nxt_actions_start_program/1,
				nxt_actions_stop_program/0,
				nxt_actions_play_sound_file/2,
				nxt_actions_stop_sound_playback/0,
				nxt_actions_play_tone/2,
				nxt_actions_get_brick_name/1,
				nxt_actions_set_brick_name/1
			       ]).

:- use_module(library(pce)).
:- use_module(threads).

:- dynamic nxt_simulator_item/3.
:- dynamic nxt_simulator_item/4.

/** <module> NXT Mindstroms - Actions Simulator.

Motors: 'A', 'B', 'C'.

@author Piotr Hoownia
@license GNU General Public License
*/


%%	nxt_actions_connection_open
%
%	Opens connection.

nxt_actions_connection_open :-
	
	% Creating main window and dialogs.
	nxt_current_robot(Alias),
	concat('PLNXT Simulator - ',Alias,WindowName),
	new(MainWindow, frame(WindowName)),
	assert(nxt_simulator_item(Alias,main_window,MainWindow)),

	% Sensors
	nxt_simulator_sensors_dialog(DialogSensors),
	send(MainWindow, append, DialogSensors),

	% Motors
	nxt_simulator_motor_dialog(DialogA, PictureA, SpeedLabelA, 'Motor A'),
	nxt_simulator_motor_dialog(DialogB, PictureB, SpeedLabelB, 'Motor B'),
	nxt_simulator_motor_dialog(DialogC, PictureC, SpeedLabelC, 'Motor C'),	
	send(MainWindow, append, DialogA),
	send(MainWindow, append, DialogC),
	send(MainWindow, append, DialogB),
	send(DialogA, below, DialogSensors),
	send(DialogB, right, DialogC),
	send(DialogB, below, DialogA),
	
	
	assert(nxt_simulator_item(Alias,picture,'A',PictureA)),
	assert(nxt_simulator_item(Alias,picture,'B',PictureB)),
	assert(nxt_simulator_item(Alias,picture,'C',PictureC)),
	assert(nxt_simulator_item(Alias,speed_label,'A',SpeedLabelA)),
	assert(nxt_simulator_item(Alias,speed_label,'B',SpeedLabelB)),
	assert(nxt_simulator_item(Alias,speed_label,'C',SpeedLabelC)),
	
	assert(nxt_simulator_item(Alias,speed,'A',0)),
	assert(nxt_simulator_item(Alias,speed,'B',0)),
	assert(nxt_simulator_item(Alias,speed,'C',0)),

	%send(DialogB, right, DialogC),
	
	%send(DialogSensors, below, MainPictures),
	%send(DialogA, below, MainPictures),

	%send(DialogC, left, DialogB),
	%send(DialogB, below, DialogA),
	%send(DialogC, below, DialogA),
	%send(DialogB, right, DialogC),
	%send(DialogB, below, DialogA),
	%send(DialogB, right, DialogC),
	%send(MainPictures, above, DialogA),
	%send(DialogSensors, above, MainPictures),

	%send(DialogC, left, DialogB),
	%send(DialogB, below, DialogA),
	%send(DialogC, below, DialogA),
	
				
	send(MainWindow, open),
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	writeln('Openning connection...'),
	write_prefix,
	writeln('Connected (simulator)!'),nl,
	mutex_unlock(nxt_simulator_mutex_device).


%%	nxt_actions_connection_close
%
%	Closes connection.

nxt_actions_connection_close :-
	nxt_current_robot(Alias),
	nxt_simulator_item(Alias,main_window,MainWindow),
	send(MainWindow,destroy),
	retractall(nxt_simulator_item(Alias,_,_)),
	retractall(nxt_simulator_item(Alias,_,_,_)),
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	writeln('Closing connection...'),
	write_prefix,
	writeln('Disconnected!'),nl,
	mutex_unlock(nxt_simulator_mutex_device).


%%	nxt_actions_motor(+Motor,+Speed,+Angle).
%
%	Rotates motor at specified speed.
%	Forward if Speed is positive, backward if Speed is negative.
%	Motor will stop, when specified revolution (Angle in degrees) is reached.
%	Motor will never stop, if Angle equals 0.

nxt_actions_motor(Motor,Speed,0) :-
	nxt_actions_motor_timer_remove(Motor),
	nxt_current_robot(Alias),
	nxt_simulator_item(Alias,picture,Motor,Picture),
	nxt_simulator_item(Alias,speed_label,Motor,SpeedLabel),
	nxt_actions_motor_move(Motor,Picture,Speed),
	retract(nxt_simulator_item(Alias,speed,Motor,_)),
	assert(nxt_simulator_item(Alias,speed,Motor,Speed)),
	send(SpeedLabel,selection,Speed).

nxt_actions_motor(Motor,Speed,Angle) :-
	Angle \= 0,
	nxt_actions_calculate_time(Speed,Angle,Time),
	timer_create(ID,Time,nxt_actions_simulator:nxt_actions_motor_stop(Motor)),
	nxt_actions_motor(Motor,Speed,0),
	nxt_actions_motor_timer_change(Motor,ID).

nxt_actions_calculate_time(Speed,Angle,Time) :-
	Speed > 0,
	nxt_simulator_time_factor(TimeFactor),
	Time is Angle/(Speed*TimeFactor).

nxt_actions_calculate_time(Speed,Angle,Time) :-
	Speed < 0,
	nxt_simulator_time_factor(TimeFactor),
	Time is -Angle/(Speed*TimeFactor).

nxt_actions_motor_move(Motor,Picture,Speed) :-
	Speed > 0,
	picture(forward,Picture),
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	write('Rotating motor '),
	write(Motor),
	write(' forward at speed: '),
	write(Speed),
	write('.'),nl,nl,
	mutex_unlock(nxt_simulator_mutex_device).

nxt_actions_motor_move(Motor,Picture,Speed) :-
	Speed < 0,
	picture(backward,Picture),
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	write('Rotating motor '),
	write(Motor),
	write(' backward at speed: '),
	write(Speed),
	write('.'),nl,nl,
	mutex_unlock(nxt_simulator_mutex_device).


%%	nxt_actions_motor_is_moving(+Motor).
%
%	Returns true if specified motor is moving. Otherwise fails.

nxt_actions_motor_is_moving(Motor) :-
	nxt_current_robot(Alias),
	nxt_simulator_item(Alias,speed,Motor,Speed),
	Speed \= 0.


%%	nxt_motor_stop(+Motor).
%
%	Stops motor.

nxt_actions_motor_stop(Motor) :-
	nxt_actions_motor_timer_remove(Motor),
	nxt_current_robot(Alias),
	nxt_simulator_item(Alias,picture,Motor,Picture),
	nxt_simulator_item(Alias,speed_label,Motor,SpeedLabel),
	picture(stopped,Picture),
	retract(nxt_simulator_item(Alias,speed,Motor,_)),
	assert(nxt_simulator_item(Alias,speed,Motor,0)),
	send(SpeedLabel,selection,0),
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	write('Motor '),
	write(Motor),
	write(' has been stopped.'),nl,nl,
	mutex_unlock(nxt_simulator_mutex_device).


%%	nxt_motor_stop(+Motor,+Brakes).
%
%	Other modules support brakes. Simulator does not, but it has to be coverd. FIXME To po angielsku jest?

nxt_actions_motor_stop(Motor,_) :-
	nxt_actions_motor_stop(Motor).


%%	nxt_actions_motor_get_speed(+Motor,-Speed).
%
%	Gets speed of the motor.

nxt_actions_motor_get_speed(Motor,Speed) :-
	nxt_current_robot(Alias),
	nxt_simulator_item(Alias,speed,Motor,Speed).


%%	nxt_actions_motor_get_rotations(+Motor,-Rot_count).
%
%	Gets rotation count of the motor.

nxt_actions_motor_get_rotations(_,_) :-
	%FIXME
	true.


%%	nxt_actions_motor_reset_position(+Motor,+Relative).
%
%	Resets motor position.

nxt_actions_motor_reset_position(_,_) :-
	%FIXME
	true.


%	nxt_actions_motor_timer_remove(+Motor).
%
%	Removes stopping timer of the Motor.

nxt_actions_motor_timer_remove(Motor) :-
	nxt_current_robot(Alias),
	nxt_simulator_item(Alias,timer,Motor,OldID),
	timer_kill(OldID),
	retractall(nxt_simulator_item(Alias,timer,Motor,_)),!.

nxt_actions_motor_timer_remove(Motor) :-
	% Needs to be retracted (Timer may be fired during the operation
	% and timer_kill/1 may fail in previous rule).
	nxt_current_robot(Alias),
       	retractall(nxt_simulator_item(Alias,timer,Motor,_)).


%	nxt_actions_motor_timer_change(+Motor,+ID).
%
%	Changes stopping timer of the Motor for a new one with the
%	specified ID.

nxt_actions_motor_timer_change(Motor,ID) :-
	nxt_current_robot(Alias),
	assert(nxt_simulator_item(Alias,timer,Motor,ID)).

		
%%	nxt_actions_touch_sensor(+Port,-Value).
%
%	Gets touch sensor reading.
%	Returns 1 if pressed, 0 otherwise.

nxt_actions_touch_sensor(_,Value) :-
	nxt_current_robot(Alias),
	nxt_simulator_item(Alias,touch,Touch),
	get(Touch,selection,Value).


%%	nxt_actions_sound_sensor(+Port,-Value).
%
%	Gets sound sensor reading.

nxt_actions_sound_sensor(_,Value) :-
	nxt_current_robot(Alias),
	nxt_simulator_item(Alias,sound,Sound),
	get(Sound,selection,Value).


%%	nxt_actions_light_sensor(+Port,-Value).
%
%	Gets light sensor reading.

nxt_actions_light_sensor(_,Value) :-
	nxt_current_robot(Alias),
	nxt_simulator_item(Alias,light,Light),
	get(Light,selection,Value).


%%	nxt_actions_light_sensor_LED(+Port,+Setting).
%
%	Sets the LED on if Setting is activate or off if passivate.

nxt_actions_light_sensor_LED(Port,Setting) :-
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	write('Light sensor LED connected to port '),
	write(Port),
	write(' has been '),
	write(Setting),
	write('d. '),nl,nl,
	mutex_unlock(nxt_simulator_mutex_device).


%%	nxt_actions_ultrasonic_sensor(+Port,-Value).
%
%	Gets ultrasonic sensor reading.

nxt_actions_ultrasonic_sensor(_,Value) :-
	nxt_current_robot(Alias),
	nxt_simulator_item(Alias,ultrasonic,Ultrasonic),
	get(Ultrasonic,selection,Value).


%%	nxt_actions_voltage(-Voltage).
%
%	Returns battery voltage in V.

nxt_actions_voltage(Voltage) :-
	Voltage is random(9),
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	write('Battery voltage (random value): '),
	write(Voltage),
	write(' V.'),nl,nl,
	mutex_unlock(nxt_simulator_mutex_device).

%%	nxt_actions_voltage_millivolt(-Voltage).
%
%	Returns battery voltage in mV.

nxt_actions_voltage_millivolt(Voltage) :-
	Voltage is random(9000),
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	write('Battery voltage (random value): '),
	write(Voltage),
	write(' mV.'),nl,nl,
	mutex_unlock(nxt_simulator_mutex_device).


%%	nxt_actions_start_program(+File)
%
%	Starts a Lego executable file on the NXT.

nxt_actions_start_program(File) :-
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	write('Executing file: '),
	write(File),nl,nl,
	mutex_unlock(nxt_simulator_mutex_device).


%%	nxt_actions_stop_program.
%
%	Stops the currently running Lego executable on the NXT.

nxt_actions_stop_program :-
	write_prefix,
	write('Stopping the current program.'),nl,nl.

%%	nxt_actions_play_sound_file(+File,+Repeat).
%
%	Plays a sound file from the NXT.
%	Files use the .rso extension.
%	The filename is not case sensitive.
%	nxt_actions_play_sound_file/2 plays sound once if Repeat is true
%	or till nxt_actions_stop_sound_playback is used if Repeat is
%	false.

nxt_actions_play_sound_file(File,true) :-
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	write('Playing in loop sound file: '),
	write(File),nl,nl,
	mutex_unlock(nxt_simulator_mutex_device).

nxt_actions_play_sound_file(File,false) :-
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	write('Playing once sound file: '),
	write(File),nl,nl,
	mutex_unlock(nxt_simulator_mutex_device).


%%	nxt_actions_stop_sound_playback.
%
%	Stops a sound file that has been playing/repeating.

nxt_actions_stop_sound_playback :-
	write_prefix,
	write('Stopping a currently played sound.'),nl,nl.


%%	nxt_actions_play_tone(Frequency,Duration).
%
%	Plays tone at specified frequency.

nxt_actions_play_tone(Frequency,Duration) :-
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	write('Playing tone at frequency: '),
	write(Frequency),
	write(' and duration: '),
	write(Duration),
	write('.'),nl,nl,
	mutex_unlock(nxt_simulator_mutex_device).


%%	nxt_actions_get_brick_name(-Name).
%
%	Gets the name of the brick.

nxt_actions_get_brick_name(Name) :-
	Name='MyBrick',
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	write('The name of the brick is: '),
	write(Name),
	write('.'),nl,nl,
	mutex_unlock(nxt_simulator_mutex_device).


%%	nxt_actions_set_brick_name(+Name)
%
%	Sets the name of the brick.

nxt_actions_set_brick_name(Name) :-
	mutex_lock(nxt_simulator_mutex_device),
	write_prefix,
	write('Setting the name of the brick to: '),
	write(Name),
	write('.'),nl,nl,
	mutex_unlock(nxt_simulator_mutex_device).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	
picture(stopped,Picture) :-
	picture_remove(Picture),
	send(Picture,display,
	     new(Text, text('Stopped')), point(3,24)),
	send(Text, font, font(times, bold, 14)),
	send(Text, colour, colour(red)),!.

picture(forward,Picture) :-
	picture_remove(Picture),
	send(Picture,display,
	     new(Arrow, path(poly, 0, chain(
					    point(30,0),
					    point(0,30),
					    point(20,30),
					    point(20,70),
					    point(40,70),
					    point(40,30),
					    point(60,30),
					    point(30,0)
					   )))),
	send(Arrow, fill_pattern, colour(green)),!.

picture(backward,Picture) :-
	picture_remove(Picture),
	send(Picture,display,
	     new(Arrow, path(poly, 0, chain(
					    point(30,70),
					    point(0,40),
					    point(20,40),
					    point(20,0),
					    point(40,0),
					    point(40,40),
					    point(60,40),
					    point(30,70)
					   )))),
	send(Arrow, fill_pattern, colour(red)),!.

picture_remove(Picture) :-
	get(Picture,member(text),PreviousPicture),
	free(PreviousPicture),!.

picture_remove(Picture) :-
	get(Picture,member(path),PreviousPicture),
	free(PreviousPicture).

nxt_simulator_sensors_dialog(Dialog) :-
	nxt_current_robot(Alias),
	new(Dialog,dialog),
	
	send(Dialog, append,
	     new(Touch, slider(touch, 0, 1, 0))),
	send(Touch, drag, @on),
	send(Touch, attribute, hor_stretch, 100),
	assert(nxt_simulator_item(Alias,touch,Touch)),
	
	send(Dialog, append,
	     new(Sound, slider(sound, 0, 100, 0))),
	send(Sound, drag, @on),
	send(Sound, attribute, hor_stretch, 100),
	assert(nxt_simulator_item(Alias,sound,Sound)),
	
	send(Dialog, append,
	     new(Light, slider(light, 0, 100, 40))),
	send(Light, drag, @on),
	send(Light, attribute, hor_stretch, 100),
	assert(nxt_simulator_item(Alias,light,Light)),
	
	send(Dialog, append,
	     new(Ultrasonic, slider(ultrasonic, 0, 255, 255))),
	send(Ultrasonic, drag, @on),
	send(Ultrasonic, attribute, hor_stretch, 100),
	assert(nxt_simulator_item(Alias,ultrasonic,Ultrasonic)).

nxt_simulator_motor_dialog(Dialog, Picture, Speed, Label) :-
	new(Dialog, dialog),
	send(Dialog, append,
	     label(label, Label)),

	send(Dialog, append, new(Picture, picture(size := size(62,73)))),
	send(Picture,display,
	     new(Text, text('Stopped')), point(3,24)),
	send(Text, font, font(times, bold, 14)),
	send(Text, colour, colour(red)),

	send(Dialog, append, new(Speed, int_item(speed))),
	send(Speed, length, 8),
	send(Speed, editable, @off),
	send(Speed, selection, 0),
	send(Speed, below, Picture).

%%%%%%%%% HELPERS %%%%%%%%%%%

%	string_to_upper(+In,-Out).
%
%	Converts string to an upper case string.

string_to_upper(In,Out) :-
	string_to_list(In,InList),
	list_to_upper(InList,OutList),
	string_to_list(Out,OutList).

list_to_upper([],[]).

list_to_upper([In|InRest],[Out|OutRest]) :-
	In > 96,
	In < 123, !,
	Out is In - 32,
	list_to_upper(InRest,OutRest).

list_to_upper([In|InRest],[In|OutRest]) :-
	list_to_upper(InRest,OutRest).


write_prefix :-
	nxt_current_robot(Alias),
	string_to_upper(Alias,Name),
	write(Name),
	get_time(T),
	convert_time(T,_,_,_,Hour,Minutes,Seconds,_),
	write(' ('),write(Hour),write(':'),write(Minutes),write(':'),write(Seconds),write('): ').
